#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <time.h>

#define BACKLOG 2
#define BUFSIZE 512

void
traite_requete (int connexion)
{
  char *buf;
  int a, b;
  int lu;

  buf = (char*) malloc (BUFSIZE);
  if (buf == NULL)
    {
      perror ("malloc");
      exit (1);
    }
  strncpy (buf, "Premier argument ?\n", BUFSIZE - 1);
  write (connexion, buf, strlen (buf) + 1);
  lu = read (connexion, buf, BUFSIZE - 1);
  if (lu <= 0)
    {
      printf ("Fin de connexion.\n");
      return;
    }
  a = atol (buf);

  strncpy (buf, "Deuxime argument ?\n", BUFSIZE - 1);
  write (connexion, buf, strlen (buf) + 1);
  lu = read (connexion, buf, BUFSIZE - 1);
  if (lu <= 0)
    {
      printf ("Fin de connexion.\n");
      return;
    }
  b = atol (buf);

  snprintf (buf, BUFSIZE - 1, "Rsultat : %d + %d = %d\n", a, b, a + b);
  write (connexion, buf, strlen (buf) + 1);
  close (connexion);
  return;
}

int
attente (int port)
{
  int sock;
  int connexion;
  struct sockaddr_in serv_addr;
  struct sockaddr_in client_addr;
  int addrlen;
  int res;
  struct hostent* host;
  int pid;

  memset (&serv_addr, 0, sizeof (serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr   = htonl (INADDR_ANY);
  serv_addr.sin_port   = htons (port);

  /* On a toutes les donnes ncessaires. On peut crer la socket */
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock == -1)
    {
      perror ("socket");
      return -1;
    }

  /* La socket est cre. On peut spcifier l'adresse/port prcis */
  res = bind (sock, (struct sockaddr*) &serv_addr, sizeof (serv_addr));
  if (res == -1)
    {
      perror ("bind");
      return -1;
    }
  
  /* On met la socket en coute */
  res = listen (sock, BACKLOG);
  if (res == -1)
    {
      perror ("listen");
      return -1;
    }

  /* On attend les connexions */
  for ( ; ; )
    {
      printf ("Attente de connexion...\n");
      connexion = accept (sock, (struct sockaddr*) &client_addr, &addrlen);
      if (connexion < 0)
	{
	  perror ("accept");
	  exit (1);
	}

      pid = fork ();
      if (pid == -1)
	{
	  perror ("fork");
	  exit (1);
	}
      if (pid == 0)
	{
	  /* On est dans le fils. */

	  /* On ferme la socket d'attente de connexion */
	  close (sock);

	  host = gethostbyaddr ((char*) &(client_addr.sin_addr.s_addr), 
				sizeof (client_addr.sin_addr),
				AF_INET);
	  if (host == NULL)
	    {
	      printf ("Connexion depuis la machine %s ...", inet_ntoa(client_addr.sin_addr.s_addr));
	    }
	  else
	    {
	      printf ("Connexion depuis la machine %s ...", host->h_name);
	    }
	  
	  traite_requete (connexion);

	  printf ("Fin de la connexion.\n");
	  close (connexion);
	  exit (0);
	}
      else
	{
	  /* Le pre ferme la socket de connexion et retourne en coute */
	  close (connexion);
	}
    }
  return -1;
}

void
sig_child (int signum)
{
  int i;
  /* Cf http://www.ee.byu.edu/unix-faq/subsection3_5_13.html */
  wait (&i);
}

int
main (int argc, char* argv[])
{
  int port;

  if (argc != 2)
    {
      fprintf (stderr, "Syntaxe : %s numero-port\n", argv[0]);
      exit (1);
    }

  signal (SIGCHLD, sig_child);
  port = atoi (argv[1]);
  printf ("Attente de connexion sur le port %d.\n", port);
  attente (port);
  exit (0);
}
